#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['hasi']

package_data = \
{'': ['*']}

install_requires = \
['itk-hasi']

extras_require = \
{'test': ['pytest >=6.2.1']}

setup(name='hasi',
      version='0.1',
      description='High-throughput Applications for Skeletal Imaging (HASI) Python content',
      author='Kitware Medical',
      author_email='matt.mccormick@kitware.com',
      url='https://github.com/KitwareMedical/HASI',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )
